@@ -5,7 +5,9 @@ from __future__ import print_function
 import re
 import os
 from os import listdir
-from os.path import isfile, join, basename
+from os.path import isfile, join, basename, dirname
+from collections import OrderedDict
+import pkg_resources
 import psycopg2
 import psycopg2.extras
 from hashlib import md5
@@ -20,25 +22,32 @@ class Upgrader:
 
     Stores the info about the upgrade in a table on the database."""
 
-    def __init__(self, pg_service, upgrades_table, directory):
+    def __init__(self, pg_service, upgrades_table, dirs, variables=None,
+                 max_version=None):
         """Constructor
 
             Parameters
             ----------
-            pg_service: string
+            pg_service: str
                 The name of the postgres service (defined in pg_service.conf)
                 related to the db
-            upgrades_table: sting
+            upgrades_table: str
                 The name of the table (int the format schema.name) where the
                 informations about the upgrades are stored
-            directory: string
-                The path of the directory where the delta files are stored
+            dirs: list(str)
+                The paths to directories where delta files are stored
+            variables: dict
+                dictionary for variables to be used in SQL deltas ( name => value )
+            max_version: str
+                Maximum (including) version to run the deltas up to.
         """
         self.pg_service = pg_service
         self.connection = psycopg2.connect("service={}".format(pg_service))
         self.cursor = self.connection.cursor()
         self.upgrades_table = upgrades_table
-        self.dir = directory
+        self.dirs = dirs
+        self.variables = variables
+        self.max_version = pkg_resources.parse_version(max_version) if max_version else None
 
     def run(self, verbose=False):
         if not self.exists_table_upgrades():
@@ -47,30 +56,28 @@ class Upgrader:
         self.__run_pre_all()
 
         deltas = self.__get_delta_files()
+        deltas = [delta for dirname in deltas for delta in deltas[dirname]]
         if not deltas:
             print('No delta files found')
-        for d in deltas:
 
+        db_version = self.current_db_version()
+
+        for d in deltas:
             if verbose:
                 print('Found delta {}, version {}, type {}'.format(
                     d.get_name(), d.get_version(), d.get_type()))
                 print('     Already applied: ', self.__is_applied(d))
-                print(
-                    '     Version greather or equal than current: ',
-                    self.__is_version_greater_or_equal_than_current(
-                        d.get_version()))
-            if (not self.__is_applied(d)) and (
-                    self.__is_version_greater_or_equal_than_current(
-                        d.get_version())):
-                print('     Applying delta {}...'.format(
-                    d.get_version()), end=' ')
+                print('     Version greather or equal than current: ',
+                      d.get_version() >= db_version)
+            if not self.__is_applied(d) and d.get_version() >= db_version and \
+                    (self.max_version is None or d.get_version() <= self.max_version):
+                print('     Applying delta {} {}...'.format(
+                    d.get_version(), d.get_name()), end=' ')
 
-                if d.get_type() in [Delta.DELTA_PRE_SQL, Delta.DELTA_SQL,
-                                  Delta.DELTA_POST_SQL]:
+                if d.get_type() & DeltaType.SQL:
                     self.__run_delta_sql(d)
                     print('OK')
-                elif d.get_type() in [Delta.DELTA_PRE_PY, Delta.DELTA_PY,
-                                    Delta.DELTA_POST_PY]:
+                elif d.get_type() & DeltaType.PYTHON:
                     self.__run_delta_py(d)
                     print('OK')
                 else:
@@ -112,21 +119,27 @@ class Upgrader:
         return self.connection.get_dsn_parameters()['user']
 
     def __get_delta_files(self):
-        """Search for delta files and return a list of Delta objects."""
-        files_in_dir = [f for f in listdir(self.dir) if isfile(
-            join(self.dir, f))]
+        """Search for delta files and return a dict of Delta objects, keyed by directory names."""
+        files = [(d, f) for d in self.dirs for f in listdir(d) if isfile(join(d, f))]
 
-        deltas = []
-        for i in files_in_dir:
-            file = join(self.dir, i)
+        deltas = OrderedDict()
+        for d, f in files:
+            file_ = join(d, f)
 
-            if not Delta.is_valid_delta_name(file):
+            if not Delta.is_valid_delta_name(file_):
                 continue
 
-            delta = Delta(file)
-            deltas.append(delta)
+            delta = Delta(file_)
 
-        return sorted(deltas, key=lambda x: (x.get_version(), x.get_type()))
+            if d not in deltas:
+                deltas[d] = []
+            deltas[d].append(delta)
+
+        # sort delta objects in each bucket
+        for d in deltas:
+            deltas[d].sort(key=lambda x: (x.get_version(), x.get_priority(), x.get_name()))
+
+        return deltas
 
     def __run_delta_sql(self, delta):
         """Execute the delta sql file on the database"""
@@ -143,32 +156,40 @@ class Upgrader:
     def __run_pre_all(self):
         """Execute the pre-all.py and pre-all.sql files if they exist"""
 
-        pre_all_py_path = os.path.join(self.dir, 'pre-all.py')
-        if os.path.isfile(pre_all_py_path):
-            print('     Applying pre-all.py...', end=' ')
-            self.__run_py_file(pre_all_py_path, 'pre-all')
-            print('OK')
+        # if the list of delta dirs is [delta1, delta2] the pre scripts of delta2 are
+        # executed before the pre scripts of delta1
 
-        pre_all_sql_path = os.path.join(self.dir, 'pre-all.sql')
-        if os.path.isfile(pre_all_sql_path):
-            print('     Applying pre-all.sql...', end=' ')
-            self.__run_sql_file(pre_all_sql_path)
-            print('OK')
+        for d in reversed(self.dirs):
+            pre_all_py_path = os.path.join(d, 'pre-all.py')
+            if os.path.isfile(pre_all_py_path):
+                print('     Applying pre-all.py...', end=' ')
+                self.__run_py_file(pre_all_py_path, 'pre-all')
+                print('OK')
+
+            pre_all_sql_path = os.path.join(d, 'pre-all.sql')
+            if os.path.isfile(pre_all_sql_path):
+                print('     Applying pre-all.sql...', end=' ')
+                self.__run_sql_file(pre_all_sql_path)
+                print('OK')
 
     def __run_post_all(self):
         """Execute the post-all.py and post-all.sql files if they exist"""
 
-        post_all_py_path = os.path.join(self.dir, 'post-all.py')
-        if os.path.isfile(post_all_py_path):
-            print('     Applying post-all.py...', end=' ')
-            self.__run_py_file(post_all_py_path, 'post-all')
-            print('OK')
+        # if the list of delta dirs is [delta1, delta2] the post scripts of delta1 are
+        # executed before the post scripts of delta2
 
-        post_all_sql_path = os.path.join(self.dir, 'post-all.sql')
-        if os.path.isfile(post_all_sql_path):
-            print('     Applying post-all.sql...', end=' ')
-            self.__run_sql_file(post_all_sql_path)
-            print('OK')
+        for d in self.dirs:
+            post_all_py_path = os.path.join(d, 'post-all.py')
+            if os.path.isfile(post_all_py_path):
+                print('     Applying post-all.py...', end=' ')
+                self.__run_py_file(post_all_py_path, 'post-all')
+                print('OK')
+
+            post_all_sql_path = os.path.join(d, 'post-all.sql')
+            if os.path.isfile(post_all_sql_path):
+                print('     Applying post-all.sql...', end=' ')
+                self.__run_sql_file(post_all_sql_path)
+                print('OK')
 
     def __run_sql_file(self, filepath):
         """Execute the sql file at the passed path
@@ -178,9 +199,13 @@ class Upgrader:
         filepath: str
             the path of the file to execute"""
 
-        delta_file = open(filepath, 'r')
-        self.cursor.execute(delta_file.read())
-        self.connection.commit()
+        with open(filepath, 'r') as delta_file:
+            sql = delta_file.read()
+            if self.variables:
+                self.cursor.execute(sql, self.variables)
+            else:
+                self.cursor.execute(sql)
+            self.connection.commit()
 
     def __run_py_file(self, filepath, module_name):
         """Execute the python file at the passed path
@@ -198,6 +223,11 @@ class Upgrader:
         delta_py = importlib.util.module_from_spec(spec)
         spec.loader.exec_module(delta_py)
 
+        # Get the python file's directory path
+        # Note: we add a separator for backward compatibility, as existing DeltaPy subclasses
+        # may assume that delta_dir ends with a separator
+        dir_ = dirname(filepath) + os.sep
+
         # Search for subclasses of DeltaPy
         for name in dir(delta_py):
             obj = getattr(delta_py, name)
@@ -205,38 +235,41 @@ class Upgrader:
                     obj, DeltaPy):
 
                 delta_py_inst = obj(
-                    self.current_db_version(), self.dir, self.pg_service,
-                    self.upgrades_table)
+                    self.current_db_version(), dir_, self.dirs, self.pg_service,
+                    self.upgrades_table, variables=self.variables)
                 delta_py_inst.run()
 
     def show_info(self):
         """Print info about found delta files and about already made upgrades"""
 
         deltas = self.__get_delta_files()
-        print('delta files in dir: ', self.dir)
+
         table = [['Version', 'Name', 'Type', 'Status']]
 
-        for delta in deltas:
-            line = [delta.get_version(), delta.get_name()]
-            if delta.get_type() == Delta.DELTA_PRE_PY:
-                line.append('pre py')
-            elif delta.get_type() == Delta.DELTA_PRE_SQL:
-                line.append('pre sql')
-            elif delta.get_type() == Delta.DELTA_PY:
-                line.append('delta py')
-            elif delta.get_type() == Delta.DELTA_SQL:
-                line.append('delta sql')
-            elif delta.get_type() == Delta.DELTA_POST_PY:
-                line.append('post py')
-            elif delta.get_type() == Delta.DELTA_POST_SQL:
-                line.append('post sql')
+        for dir_ in deltas:
+            print('delta files in dir: ', dir_)
 
-            if self.__is_applied(delta):
-                line.append('Applied')
-            else:
-                line.append('Pending')
+            for delta in deltas[dir_]:
+                line = [str(delta.get_version()), delta.get_name()]
+                if delta.get_type() == DeltaType.PRE_PYTHON:
+                    line.append('pre py')
+                elif delta.get_type() == DeltaType.PRE_SQL:
+                    line.append('pre sql')
+                elif delta.get_type() == DeltaType.PYTHON:
+                    line.append('delta py')
+                elif delta.get_type() == DeltaType.SQL:
+                    line.append('delta sql')
+                elif delta.get_type() == DeltaType.POST_PYTHON:
+                    line.append('post py')
+                elif delta.get_type() == DeltaType.POST_SQL:
+                    line.append('post sql')
 
-            table.append(line)
+                if self.__is_applied(delta):
+                    line.append('Applied')
+                else:
+                    line.append('Pending')
+
+                table.append(line)
 
         self.__print_table(table)
 
@@ -264,17 +297,17 @@ class Upgrader:
             delta_type = i[2]
             if delta_type == 0:
                 line.append('baseline')
-            elif delta_type == Delta.DELTA_PRE_PY:
+            elif delta_type == DeltaType.PRE_PYTHON:
                 line.append('pre py')
-            elif delta_type == Delta.DELTA_PRE_SQL:
+            elif delta_type == DeltaType.PRE_SQL:
                 line.append('pre sql')
-            elif delta_type == Delta.DELTA_PY:
+            elif delta_type == DeltaType.PYTHON:
                 line.append('delta py')
-            elif delta_type == Delta.DELTA_SQL:
+            elif delta_type == DeltaType.SQL:
                 line.append('delta sql')
-            elif delta_type == Delta.DELTA_POST_PY:
+            elif delta_type == DeltaType.POST_PYTHON:
                 line.append('post py')
-            elif delta_type == Delta.DELTA_POST_SQL:
+            elif delta_type == DeltaType.POST_SQL:
                 line.append('post sql')
 
             line.append(str(i[3]))
@@ -377,7 +410,7 @@ class Upgrader:
         query = """CREATE TABLE IF NOT EXISTS {}
                 (
                 id serial NOT NULL,
-                version character varying(50),
+                version character varying(50) NOT NULL,
                 description character varying(200) NOT NULL,
                 type integer NOT NULL,
                 script character varying(1000) NOT NULL,
@@ -386,7 +419,8 @@ class Upgrader:
                 installed_on timestamp without time zone NOT NULL DEFAULT now(),
                 execution_time integer NOT NULL,
                 success boolean NOT NULL,
-                PRIMARY KEY (id)
+                PRIMARY KEY (id),
+                EXCLUDE (version WITH =) WHERE (type = 0)
                 )
         """.format(self.upgrades_table)
 
@@ -428,27 +462,6 @@ class Upgrader:
         self.cursor.execute(query)
         self.connection.commit()
 
-    def __is_version_greater_or_equal_than_current(self, version):
-        """Read the upgrades information table and return if the passed
-        version is greater or equal than the current db version
-
-        Parameters
-        ----------
-        version: str
-            the version to be compared with the current db version
-
-        Returns
-        -------
-        bool
-            True id the passed version is greater or equal to the current db
-            version,
-            False otherwise
-        """
-
-        if version >= self.current_db_version():
-            return True
-        return False
-
     def current_db_version(self):
         """Read the upgrades information table and return the current db
         version
@@ -465,19 +478,26 @@ class Upgrader:
 
         self.cursor.execute(query)
 
-        return self.cursor.fetchone()[0]
+        return pkg_resources.parse_version(self.cursor.fetchone()[0])
+
+
+class DeltaType:
+    PRE = 1
+    POST = 2
+
+    PYTHON = 4
+    SQL = 8
+
+    PRE_PYTHON = PRE | PYTHON
+    PRE_SQL = PRE | SQL
+
+    POST_PYTHON = POST | PYTHON
+    POST_SQL = POST | SQL
 
 
 class Delta:
     """This class represent a delta file."""
 
-    DELTA_PRE_PY = 0
-    DELTA_PRE_SQL = 1
-    DELTA_PY = 2
-    DELTA_SQL = 3
-    DELTA_POST_PY = 4
-    DELTA_POST_SQL = 5
-
     FILENAME_PATTERN = (
         r"^(delta_)(\d+\.\d+\.\d+)(_*)(\w*)\."
         r"(pre\.sql|sql|post\.sql|pre\.py|py|post\.py)$")
@@ -513,9 +533,12 @@ class Delta:
         pattern = re.compile(self.FILENAME_PATTERN)
         self.match = re.match(pattern, filename)
 
+    def __repr__(self):
+        return '<file: {f} (v: {v}, n: {n}>'.format(f=self.file, v=self.get_version(), n=self.get_name())
+
     def get_version(self):
         """Return the version of the delta file."""
-        return self.match.group(2)
+        return pkg_resources.parse_version(self.match.group(2))
 
     def get_name(self):
         """Return the name (description) of the delta file."""
@@ -523,7 +546,9 @@ class Delta:
 
     def get_checksum(self):
         """Return the md5 checksum of the delta file."""
-        return md5(open(self.file, 'rb').read()).hexdigest()
+        with open(self.file, 'rb') as f:
+            cs = md5(f.read()).hexdigest()
+        return cs
 
     def get_type(self):
         """Return the type of the delta file.
@@ -536,17 +561,32 @@ class Delta:
         ext = self.match.group(5)
 
         if ext == 'pre.py':
-            return Delta.DELTA_PRE_PY
+            return DeltaType.PRE_PYTHON
         elif ext == 'pre.sql':
-            return Delta.DELTA_PRE_SQL
+            return DeltaType.PRE_SQL
         elif ext == 'py':
-            return Delta.DELTA_PY
+            return DeltaType.PYTHON
         elif ext == 'sql':
-            return Delta.DELTA_SQL
+            return DeltaType.SQL
         elif ext == 'post.py':
-            return Delta.DELTA_POST_PY
+            return DeltaType.POST_PYTHON
         elif ext == 'post.sql':
-            return Delta.DELTA_POST_SQL
+            return DeltaType.POST_SQL
+
+    def get_priority(self) -> int:
+        """Return the priority of the file from 1 (pre) to 3 (post)
+
+        Returns
+        -------
+        type: int
+        """
+        dtype = self.get_type()
+        if dtype & DeltaType.PRE:
+            return 1
+        elif dtype & DeltaType.POST:
+            return 3
+        else:
+            return 2
 
     def get_file(self):
         return self.file
